/*******************************************************************************
* Copyright (c) 2015, Maximilian Fickert
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of the author may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

#pragma once

#include <queue>
#include <utility>
#include <vector>

template <class K, class V> class hybrid_queue {

public:
	using value_type = std::pair<K, V>;

	hybrid_queue()
			: bucket(),
			  pqueue([](const value_type &lhs, const value_type &rhs) {
				  return lhs.first > rhs.first;
			  }) {}

	void reserve_bucket(size_t new_cap) const { bucket.reserve(new_cap); }

	void push(const value_type &value) {
		const auto &key = value.first;
		if (key > min_key) {
			pqueue.push(value);
			return;
		}
		if (key == min_key) {
			bucket.push_back(value);
			return;
		}
		move_bucket_to_queue();
		bucket.push_back(value);
		min_key = key;
	}

	auto top() -> value_type {
		assert(!empty());
		if (bucket.empty())
			return pqueue.top();
#ifndef NDEBUG
		if (!pqueue.empty())
			for (const auto &entry : bucket)
				assert(entry.first <= pqueue.top().first);
#endif
		return bucket.back();
	}

	void pop() {
		assert(!empty());
		if (bucket.empty()) {
			min_key = pqueue.top().first;
			pqueue.pop();
		} else {
			bucket.pop_back();
		}
	}

	auto empty() const -> bool { return bucket.empty() && pqueue.empty(); }

	void clear() {
		bucket.clear();
		pqueue.clear();
	}

private:
	void move_bucket_to_queue() {
		for (const auto &entry : bucket)
			pqueue.push(entry);
		bucket.clear();
	}

	K min_key;
	std::vector<value_type> bucket;
	std::priority_queue<value_type, std::vector<value_type>,
						bool (*)(const value_type &, const value_type &rhs)>
			pqueue;
};
