/*******************************************************************************
* Copyright (c) 2015, Maximilian Fickert, Andreas Karrenbauer
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * The name of the author may not be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

#include "graph.h"
#include "min_cost_flow_dual_ascent.h"
#include "min_cost_flow_dual_ascent_default.h"

#include <chrono>
#include <fstream>
#include <iostream>

int main() {
	std::cout << "Enter filepath to input graph.. " << std::endl;
	std::string filename;
	std::cin >> filename;

	// check if file exists
	if (!std::ifstream(filename)) {
		std::cerr << "File not found: " << filename << std::endl;
		exit(1);
	}

	using RationalType = long long int;
	using IntegerType = long long int;

	Graph<IntegerType, RationalType> G(filename);
#ifdef RESTORE_BALANCED_NODES
	Network<decltype(G), IntegerType, RationalType,
			DualAscentNodeData<IntegerType, RationalType>,
			DualAscentRBNArcData<RationalType>>
			N(G, filename);
#else
	Network<decltype(G), IntegerType, RationalType,
			DualAscentNodeData<IntegerType, RationalType>,
			BasicArcData<RationalType>>
			N(G, filename);
#endif

	cout << "Number of Edges: " << N.G.no_of_edges << endl;
	cout << "Number of Nodes: " << N.G.no_of_vertices << endl;

	auto start = std::chrono::steady_clock::now();

	dual_ascent<decltype(N), DefaultNodeIterator, DefaultEdgeIterator,
				DefaultNodeAccessor, DefaultEdgeAccessor,
				DefaultIncidentEdgesIterator>(
			N, DefaultNodeIterator(1),
			DefaultNodeIterator(N.G.no_of_vertices + 1));

	auto elapsed_time = std::chrono::duration<double>(
			std::chrono::steady_clock::now() - start);
	std::cout << "total time: " << elapsed_time.count() << "s" << std::endl;

	// check flow conservation
	for (auto v = 1u; v < N.G.no_of_vertices; ++v) {
		auto sum_v = static_cast<decltype(N.arcdata.front().xlower)>(0);
		for (auto edge : N.G.incident_edges[v]) {
			if (edge > 0) {
				sum_v -= N.arcdata[edge].xlower;
			} else {
				sum_v += N.arcdata[-edge].xlower;
			}
		}
		if (!(sum_v == N.nodedata[v].demand)) {
			std::cerr << "flow conservation check failed" << std::endl;
			exit(1);
		}
	}

	// check complementary slackness et al
	for (auto a = 1u; a < N.G.no_of_edges; ++a) {
		const auto &arcdata = N.arcdata[a];
		const auto &i = N.G.tails[a];
		const auto &j = N.G.heads[a];

		// check primal feasibility
		if (!(arcdata.xlower <= arcdata.capacity)) {
			std::cerr << "capacity constraints check failed" << std::endl;
			exit(1);
		}

#ifdef INFINITE_CAPACITY
		// check dual feasibility
		if (!(arcdata.capacity != INFINITE_CAPACITY ||
			  arcdata.cost + N.nodedata[i].potential -
							  N.nodedata[j].potential >=
					  0)) {
			std::cerr << "dual feasibility check failed" << std::endl;
			exit(1);
		}
#endif

		// check complementary slackness
		if (!(arcdata.cost + N.nodedata[i].potential -
							  N.nodedata[j].potential <=
					  0 ||
			  arcdata.xlower == 0)) {
			std::cerr << "complementary slackness check 1 failed" << std::endl;
			exit(1);
		}
		if (!(arcdata.cost + N.nodedata[i].potential -
							  N.nodedata[j].potential >=
					  0 ||
			  arcdata.xlower == arcdata.capacity)) {
			std::cerr << "complementary slackness check 2 failed" << std::endl;
			exit(1);
		}
		if (!(!(arcdata.xlower > 0 && arcdata.xlower < arcdata.capacity) ||
			  arcdata.cost + N.nodedata[i].potential -
							  N.nodedata[j].potential ==
					  0)) {
			std::cerr << "complementary slackness check 3 failed" << std::endl;
			exit(1);
		}
	}

	return 0;
}
